#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <utime.h>
#include <errno.h>
#include "embfeatures.h"

#ifdef __GLIBC__
#define __NO_STAT64
#endif

#ifdef TAR_SUPPORT_SYMBOLIC_UID
#include <pwd.h>
#include <grp.h>

#define EMBEDDED
#include "openreadclose.c"
#include "split.c"
#define DONT_NEED_PASSWD
#define DONT_NEED_GECOS
#define DONT_NEED_DIR
#define DONT_NEED_SHELL
#include "mygetpwnam.c"
#define DONT_NEED_MEMBERS
#include "mygetgrnam.c"
#endif

#include "fmt_str.c"
#include "fmt_ulong.c"
#define DONT_NEED_BUFFER_1
#include "buffer.c"

char buffer_2_space[BUFFER_INSIZE];
static buffer it2 = BUFFER_INIT(write,2,buffer_2_space,sizeof buffer_2_space);
buffer *buffer_2 = &it2;

int fmt_pad(char *dest,char *src,char pad,int padto,int len) {
  int i;
  register char *d;
  if (!(d=dest))
    return len<padto?padto:len;
  while (len<padto) { *d=pad; ++d; --padto; }
  for (i=0; i<len; ++i) { *d=src[i]; ++d; }
  return d-dest;
}

void die(const char *message) {
  write(2,message,strlen(message));
  close(0);
  exit(1);
}

void couldnot(const char *what,const char *name) {
  buffer_puts(buffer_2,"Could not ");
  buffer_puts(buffer_2,what);
  buffer_puts(buffer_2," \"");
  buffer_puts(buffer_2,name);
  buffer_puts(buffer_2,"\": ");
  buffer_puts(buffer_2,strerror(errno));
  buffer_putc(buffer_2,'\n');
  buffer_flush(buffer_2);
}

struct posix_header
{				/* byte offset */
  char name[100];		/*   0 */
  char mode[8];			/* 100 */
  char uid[8];			/* 108 */
  char gid[8];			/* 116 */
  char size[12];		/* 124 */
  char mtime[12];		/* 136 */
  char chksum[8];		/* 148 */
  char typeflag;		/* 156 */
  char linkname[100];		/* 157 */
  char magic[6];		/* 257 */
  char version[2];		/* 263 */
  char uname[32];		/* 265 */
  char gname[32];		/* 297 */
  char devmajor[8];		/* 329 */
  char devminor[8];		/* 337 */
  char prefix[155];		/* 345 */
				/* 500 */
};

void __m2s(int i, char *s) {
  if (i&4) s[0]='r'; else s[0]='-';
  if (i&2) s[1]='w'; else s[1]='-';
  if (i&1) s[2]='x'; else s[2]='-';
}

void mode2str(mode_t m,char *s) {
  if (S_ISREG(m)) s[0]='-';
  else if (S_ISCHR(m)) s[0]='c';
  else if (S_ISDIR(m)) s[0]='d';
  else if (S_ISBLK(m)) s[0]='b';
  else if (S_ISFIFO(m)) s[0]='p';
  else if (S_ISSOCK(m)) s[0]='s';
  else if (S_ISLNK(m)) s[0]='l';
  else s[0]='?';
  __m2s(m>>6,s+1);
  __m2s(m>>3,s+4);
  __m2s(m,s+7);
  if (m&S_ISUID) s[1]='s';
  if (m&S_ISGID) s[4]='s';
}

void time2str(time_t t,char *s) {
  struct tm *T=localtime(&t);
  char buf[50];
  /* 1999-09-11 01:10:32 */
  s+=fmt_pad(s,buf,'0',4,fmt_ulong(buf,T->tm_year+1900));
  *s='-'; ++s;
  s+=fmt_pad(s,buf,'0',2,fmt_ulong(buf,T->tm_mon+1));
  *s='-'; ++s;
  s+=fmt_pad(s,buf,'0',2,fmt_ulong(buf,T->tm_mday));
  *s=' '; ++s;
  s+=fmt_pad(s,buf,'0',2,fmt_ulong(buf,T->tm_hour));
  *s='-'; ++s;
  s+=fmt_pad(s,buf,'0',2,fmt_ulong(buf,T->tm_min));
  *s='-'; ++s;
  s+=fmt_pad(s,buf,'0',2,fmt_ulong(buf,T->tm_sec));
  *s=' '; ++s;
  *s=0;
/* int fmt_pad(char *dest,char *src,char pad,int padto,int len) */
/*  snprintf(s,22,"%04d-%02d-%02d %02d-%02d-%02d ",
	   T->tm_year+1900,T->tm_mon+1,T->tm_mday,
	   T->tm_hour,T->tm_min,T->tm_sec); */
}

int piperead(int fd,void *buf,int len) {
  int x=0;
  do {
    int tmp=read(fd,buf,len);
    if (tmp<0) return -1;
    if (tmp==0) return x;
    x+=tmp;
    len-=tmp;
    buf+=tmp;
  } while (len);
  return x;
}

void domkdir(char *path) {
  char *tmp=path;
  while (*tmp) {
    char old;
    for (; *tmp && *tmp!='/'; ++tmp) ;
    if ((old=*tmp)) {
      if (tmp[1]==0) break;
      *tmp=0;
/*      printf("mkdir %s\n",path); */
      mkdir(path,0755);
      *tmp=old;
      ++tmp;
    } else
      break;
  }
}

/* el-cheapo tar that can only create and extract archives */
int main(int argc,char *argv[]) {
  int myuid=geteuid();
  int compress=0;
  int unlink_first=0;
  int verbose=0;
  const char *file=0;
  char **filez;
  int tarfd;
  int filefd;
  char buf[10240];
  int len;
  enum {
    NOOP,
    CREATE,
    EXTRACT,
    TEST
  } todo=NOOP;
  if (argc<2)
    die("usage: tar czvf file.tar.gz filename...\n"
	"       tar xzvUf file.tar.gz\n");
  switch (*argv[1]) {
  case 'c': todo=CREATE; break;
  case 't': todo=TEST; break;
  case 'x': todo=EXTRACT; break;
  default:
	    die("unknown command, expected [ctx]\n");
  }
  {
    char *tmp=argv[1];
    while ((tmp=strchr(tmp,'v'))) {
      ++verbose;
      ++tmp;
    }
  }
  if (strchr(argv[1],'z')) compress=1;
  if (strchr(argv[1],'I')) compress=2;
  if (strchr(argv[1],'U')) unlink_first=1;
  filez=argv+2;
  if (strchr(argv[1],'f')) file=*filez++;
  switch (todo) {
  case TEST:
    ++verbose;
  case EXTRACT:
    close(0);
    tarfd=open(file,O_RDONLY);
    if (tarfd<0) die("could not open tar file\n");
    if (tarfd!=0)
      if (dup2(tarfd,0)<0) die("dup2 failed\n");
    tarfd=0;

    if (compress) {
      char *argv[]={"gzip","-dc",0};
      int fds[2];
      if (compress==2) argv[0]="bzip2";
      if (pipe(fds)) die("pipe failed\n");
      switch (fork()) {
      case -1:
	die("could not fork to exec gzip/bzip2");
      case 0:	/* child */
	if (dup2(fds[1],1)<0) die("dup2 failed\n");
	execvp(argv[0],argv);
	die("could not execvp gzip/bzip2");
      default:
	if (dup2(fds[0],0)<0) die("dup2 failed\n");
      }
    } else {
      tarfd=open(file,O_RDONLY);
      if (tarfd<0) die("could not open tar file\n");
      if (dup2(tarfd,0)<0) die("dup2 failed\n");
    }

    umask(0);
    do {
#if 0
      {
	char tmp[100];
	strcpy(tmp,"ofs ");
	tmp[4+__ltostr(tmp+4,90,lseek(0,0,SEEK_CUR),10,0)]=0;
	write(2,tmp,strlen(tmp)); write(2,"\n",1);
      }
#endif
      len=piperead(0,buf,512);
      if (len<512) {
	if (len==0) break;
	die("unexpected end of tar file\n");
      }
      if (!buf[0]) break;
      {
	struct posix_header *ph=(struct posix_header*)buf;
#ifdef __NO_STAT64
	struct stat ss;
#else
	struct stat64 ss;
#endif
	if (memcmp(ph->magic,"ustar",5))
	  die("wrong magic -- corrupt tar file\n");
	if (ph->name[99])
	  die("this does not look like a tar file to me\n");
	ph->uname[31]=0; ph->gname[31]=0;
	memset(&ss,0,sizeof(ss));
	ss.st_size=strtoull(ph->size,0,8);
	ss.st_mode=strtoul(ph->mode,0,8);
	ss.st_dev=0;
	ss.st_uid=strtoul(ph->uid,0,8);
	ss.st_gid=strtoul(ph->gid,0,8);
#ifdef TAR_SUPPORT_SYMBOLIC_UID
	{
	  struct passwd *pwd;
	  struct group *grp;
	  if ((pwd=getpwnam(ph->uname)))
	    ss.st_uid=pwd->pw_uid;
	  if ((grp=getgrnam(ph->gname)))
	    ss.st_gid=grp->gr_gid;
	}
#endif
	{
	  int maj,min;
	  maj=strtoul(ph->devmajor,0,8);
	  min=strtoul(ph->devminor,0,8);
	  ss.st_dev=makedev(maj,min);
	}
	ss.st_mtime=strtoul(ph->mtime,0,8);
	if (verbose>=1) {
	  if (verbose>=2) {
	    char tmp[30];
	    static int muglen=6;
	    static int mslen=5;
	    int uglen,slen;
	    mode2str(ss.st_mode,tmp);
	    buffer_put(buffer_2,tmp,10);
	    buffer_putc(buffer_2,' ');

	    {
	      char tmp[200];
	      if (*ph->uname)
		uglen=fmt_str(tmp,ph->uname);
	      else
		uglen=fmt_ulong(tmp,ss.st_uid);
	      tmp[uglen]='/'; ++uglen;
	      if (*ph->uname)
		uglen+=fmt_str(tmp+uglen,ph->gname);
	      else
		uglen+=fmt_ulong(tmp+uglen,ss.st_gid);
	      if (uglen>muglen) muglen=uglen;
	      buffer_put(buffer_2,tmp,uglen);
	    }
/*	    tmp+=snprintf(tmp,BUFSIZE-(tmp-Buf),"%s/%s ",ph->uname,ph->gname); */
	    while (uglen<muglen) {
	      ++uglen;
	      buffer_putc(buffer_2,' ');
	    }
	    buffer_putc(buffer_2,' ');

	    {
	      char numbuf[100];
	      if (S_ISCHR(ss.st_mode) || S_ISBLK(ss.st_mode)) {
		slen=fmt_ulong(numbuf,major(ss.st_dev));
		numbuf[slen]=','; ++slen;
		slen+=fmt_ulong(numbuf,minor(ss.st_dev));
		/*slen=snprintf(numbuf,100,"%u,%u",major(ss.st_dev),minor(ss.st_dev)); */
	      } else
		slen=fmt_ulong(numbuf,ss.st_size);
		/*slen=snprintf(numbuf,100,"%lu",ss.st_size); */
	      if (slen>mslen) mslen=slen;
	      {
		char numbuf2[100];
		buffer_put(buffer_2,numbuf2,fmt_pad(numbuf2,numbuf,' ',mslen,slen));
	      }
	      /*buffer_put(buffer_2,numbuf,
	      memset(tmp,' ',mslen);
	      strcpy(tmp+mslen-slen,numbuf);
	      tmp+=mslen; */
	    }
	    buffer_putc(buffer_2,' ');
	    /**tmp++=' '; */

	    time2str(ss.st_mtime,tmp);
	    buffer_puts(buffer_2,tmp);
	    /*tmp+=strlen(tmp); */
	    buffer_puts(buffer_2,ph->name);
	    /*strncat(tmp,ph->name,BUFSIZE-(tmp-Buf)-2); */
	    if (S_ISLNK(ss.st_mode)) {
	      buffer_puts(buffer_2," -> ");
	      buffer_puts(buffer_2,ph->linkname);
	      /*strlcat(tmp," -> ",BUFSIZE);
	      strlcat(tmp,ph->linkname,BUFSIZE); */
	    }
	    buffer_putc(buffer_2,'\n');
	    /*strlcat(tmp,"\n",BUFSIZE);*/
	    buffer_flush(buffer_2);

	    /*write(2,Buf,strlen(Buf));*/
	  } else {
	    buffer_puts(buffer_2,buf);
	    buffer_putc(buffer_2,'\n');
	    buffer_flush(buffer_2);
	    /*write(2,buf,strlen(buf)); write(2,"\n",1);*/
	  }
	}
	if (todo==EXTRACT) {
	  filefd=-1;
	  if (unlink_first)
	    if (unlink(ph->name))
	      couldnot("delete file",ph->name);
	  domkdir(ph->name);
	  if (S_ISLNK(ss.st_mode)) {
	    if (symlink(ph->linkname,ph->name))
	      couldnot("create symlink",ph->name);
	  } else if (S_ISCHR(ss.st_mode) || S_ISBLK(ss.st_mode) || S_ISFIFO(ss.st_mode)) {
	    if (mknod(ph->name,ss.st_mode,ss.st_dev))
	      couldnot("mknod",ph->name);
	    if (myuid==0)
	      chown(ph->name,ss.st_uid,ss.st_gid);
	  } else if (S_ISDIR(ss.st_mode)) {
	    if (mkdir(ph->name,ss.st_mode))
	      couldnot("create directory",ph->name);
	  } else {
	    filefd=open(ph->name,O_WRONLY|O_CREAT|O_TRUNC,ss.st_mode);
	    if (filefd<0)
	      couldnot("create file",ph->name);
	  }
	} else filefd=-1;
	if (ss.st_size>0) {
	  long long realsize=(ss.st_size+511)&(~511);
	  long long towrite=ss.st_size;
	  while (realsize>0) {
	    long long foo=(realsize>1024?1024:realsize);
	    char tmp[1024];
	    int len=piperead(0,tmp,foo);
	    int need=len>towrite?towrite:len;
	    if (len<foo) die("short read\n");
	    if (filefd>=0) {
	      if (write(filefd,tmp,need)<need)
		die("short write\n");
	      else
		towrite-=need;
	    }
	    realsize-=foo;
	  }
	  if (filefd>=0)
	    close(filefd);
	}
	if (filefd>=0 && !S_ISLNK(ss.st_mode)) {
	  struct utimbuf ut;
	  ut.actime=ut.modtime=ss.st_mtime;
	  if (utime(ph->name,&ut))
	    couldnot("set utime of ",ph->name);
	}
      }
    } while (1);
  }
  return 0;
}
